#version 120

/*
 _______ _________ _______  _______  _ 
(  ____ \\__   __/(  ___  )(  ____ )( )
| (    \/   ) (   | (   ) || (    )|| |
| (_____    | |   | |   | || (____)|| |
(_____  )   | |   | |   | ||  _____)| |
      ) |   | |   | |   | || (      (_)
/\____) |   | |   | (___) || )       _ 
\_______)   )_(   (_______)|/       (_)

Do not modify this code until you have read the LICENSE.txt contained in the root directory of this shaderpack!

*/



////////////////////////////////////////////////////ADJUSTABLE VARIABLES/////////////////////////////////////////////////////////

#define NORMAL_MAP_MAX_ANGLE 1.0f   		//The higher the value, the more extreme per-pixel normal mapping (bump mapping) will be.
//#define TEXTURE_FILTERING
#define TILE_RESOLUTION 128

#define PARALLAX

///////////////////////////////////////////////////END OF ADJUSTABLE VARIABLES///////////////////////////////////////////////////

/* DRAWBUFFERS:0123 */

uniform sampler2D texture;
uniform sampler2D lightmap;
uniform sampler2D normals;
uniform sampler2D specular;
uniform sampler2D noisetex;
//uniform float rainStrength;
uniform float rainStrength;
uniform float frameTimeCounter;
uniform vec3 sunPosition;

uniform float near;
uniform float far;
uniform float viewWidth;
uniform float viewHeight;
uniform float aspectRatio;

varying vec4 color;
varying vec4 texcoord;
varying vec4 lmcoord;
varying vec3 worldPosition;
varying vec4 vertexPos;
varying mat3 tbnMatrix;

varying vec3 normal;
varying vec3 tangent;
varying vec3 binormal;

varying float materialIDs;

varying float distance;
varying float idCheck;

const int GL_LINEAR = 9729;
const int GL_EXP = 2048;

const float bump_distance = 78.0f;
const float fademult = 0.1f;

vec4 cubic(float x)
{
    float x2 = x * x;
    float x3 = x2 * x;
    vec4 w;
    w.x =   -x3 + 3*x2 - 3*x + 1;
    w.y =  3*x3 - 6*x2       + 4;
    w.z = -3*x3 + 3*x2 + 3*x + 1;
    w.w =  x3;
    return w / 6.f;
}

vec4 BicubicTexture(in sampler2D tex, in vec2 coord)
{
	int resolution = 64;

	coord *= resolution;

	float fx = fract(coord.x);
    float fy = fract(coord.y);
    coord.x -= fx;
    coord.y -= fy;

    vec4 xcubic = cubic(fx);
    vec4 ycubic = cubic(fy);

    vec4 c = vec4(coord.x - 0.5, coord.x + 1.5, coord.y - 0.5, coord.y + 1.5);
    vec4 s = vec4(xcubic.x + xcubic.y, xcubic.z + xcubic.w, ycubic.x + ycubic.y, ycubic.z + ycubic.w);
    vec4 offset = c + vec4(xcubic.y, xcubic.w, ycubic.y, ycubic.w) / s;

    vec4 sample0 = texture2D(tex, vec2(offset.x, offset.z) / resolution);
    vec4 sample1 = texture2D(tex, vec2(offset.y, offset.z) / resolution);
    vec4 sample2 = texture2D(tex, vec2(offset.x, offset.w) / resolution);
    vec4 sample3 = texture2D(tex, vec2(offset.y, offset.w) / resolution);

    float sx = s.x / (s.x + s.y);
    float sy = s.z / (s.z + s.w);

    return mix( mix(sample3, sample2, sx), mix(sample1, sample0, sx), sy);
}


vec2 OffsetCoord(in vec2 coord, in vec2 offset, in int level)
{
	int tileResolution = TILE_RESOLUTION;
	ivec2 atlasTiles = ivec2(32, 16);
	ivec2 atlasResolution = tileResolution * atlasTiles;

	coord *= atlasResolution;

	vec2 offsetCoord = coord + mod(offset.xy * atlasResolution, vec2(tileResolution));

	vec2 minCoord = vec2(coord.x - mod(coord.x, tileResolution), coord.y - mod(coord.y, tileResolution));
	vec2 maxCoord = minCoord + tileResolution;

	if (offsetCoord.x > maxCoord.x) {
		offsetCoord.x -= tileResolution;
	} else if (offsetCoord.x < minCoord.x) {
		offsetCoord.x += tileResolution;
	}

	if (offsetCoord.y > maxCoord.y) {
		offsetCoord.y -= tileResolution;
	} else if (offsetCoord.y < minCoord.y) {
		offsetCoord.y += tileResolution;
	}

	offsetCoord /= atlasResolution;

	return offsetCoord;
}


vec4 TextureSmooth(in sampler2D tex, in vec2 coord, in int level, in bool alphaCheck)
{

	#ifdef TEXTURE_FILTERING
	level = 0;
	float scale = pow(2.0f, float(level));
	vec2 atlasTiles = vec2(32.0f, 16.0f);
	float tileResolution = TILE_RESOLUTION;
		  tileResolution /= scale;
	vec2 res = atlasTiles * tileResolution;
	coord = coord * res;
	vec2 i = floor(coord);
	vec2 f = fract(coord);
	f = f * f * (3.0f - 2.0f * f);

	//i -= vec2(0.5f);

	vec2 minI = vec2(i.x - mod(i.x, tileResolution), i.y - mod(i.y, tileResolution));
	vec2 maxI = vec2(minI.x + tileResolution, minI.y + tileResolution);
		 maxI -= 1.0f;

	

	vec2 icoordCenter = i / res;
	vec2 icoordRight 		= (i + vec2(1.0f, 0.0f));

		if (icoordRight.x < minI.x)
			icoordRight.x += tileResolution;
		else if (icoordRight.x > maxI.x)
			icoordRight.x -= tileResolution;

		if (icoordRight.y < minI.y)
			icoordRight.y += tileResolution;
		else if (icoordRight.y > maxI.y)
			icoordRight.y -= tileResolution;

		icoordRight /= res;


	vec2 icoordUp	 		= (i + vec2(0.0f, 1.0f));

		if (icoordUp.x < minI.x)
			icoordUp.x += tileResolution;
		else if (icoordUp.x > maxI.x)
			icoordUp.x -= tileResolution;

		if (icoordUp.y < minI.y)
			icoordUp.y += tileResolution;
		else if (icoordUp.y > maxI.y)
			icoordUp.y -= tileResolution;

		icoordUp /= res;


	vec2 icoordUpRight	 	= (i + vec2(1.0f, 1.0f));

		if (icoordUpRight.x < minI.x)
			icoordUpRight.x += tileResolution;
		else if (icoordUpRight.x > maxI.x)
			icoordUpRight.x -= tileResolution;

		if (icoordUpRight.y < minI.y)
			icoordUpRight.y += tileResolution;
		else if (icoordUpRight.y > maxI.y)
			icoordUpRight.y -= tileResolution;

		icoordUpRight /= res;



	vec4 texCenter 	= texture2DLod(tex, icoordCenter, 	level);
	vec4 texRight 	= texture2DLod(tex, icoordRight, 	level);
	vec4 texUp 		= texture2DLod(tex, icoordUp, 		level);
	vec4 texUpRight	= texture2DLod(tex, icoordUpRight,  level);

	if (alphaCheck)
	{
		if (texCenter.a < 0.5f && texCenter.r + texCenter.g + texCenter.b < 0.1f)
			texCenter.rgb = max(texRight.rgb, max(texUp.rgb, texUpRight.rgb));
		else if (texCenter.a < 0.5f && texCenter.r + texCenter.g + texCenter.b > 0.1f)
			texCenter.rgb = min(texRight.rgb, min(texUp.rgb, texUpRight.rgb));

		if (texUp.a < 0.5f && texUp.r + texUp.g + texUp.b < 0.1f)
			texUp.rgb = max(texCenter.rgb, max(texRight.rgb, texUpRight.rgb));
		else if (texUp.a < 0.5f && texUp.r + texUp.g + texUp.b > 0.1f)
			texUp.rgb = min(texCenter.rgb, min(texRight.rgb, texUpRight.rgb));

		if (texRight.a < 0.5f && texRight.r + texRight.g + texRight.b < 0.1f)
			texRight.rgb = max(texCenter.rgb, max(texUpRight.rgb, texUp.rgb));
		else if (texRight.a < 0.5f && texRight.r + texRight.g + texRight.b > 0.1f)
			texRight.rgb = min(texCenter.rgb, min(texUpRight.rgb, texUp.rgb));

		if (texUpRight.a < 0.5f && texUpRight.r + texUpRight.g + texUpRight.b < 0.1f)
			texUpRight.rgb = max(texCenter.rgb, max(texUp.rgb, texRight.rgb));
		else if (texUpRight.a < 0.5f && texUpRight.r + texUpRight.g + texUpRight.b > 0.1f)
			texUpRight.rgb = min(texCenter.rgb, min(texUp.rgb, texRight.rgb));
	}

	texCenter = mix(texCenter, texUp, vec4(f.y));
	texRight  = mix(texRight, texUpRight, vec4(f.y));

	vec4 result = mix(texCenter, texRight, vec4(f.x));
	return result;
	#else
	return texture2D(tex, coord, 0);
	#endif
}

vec3 Get3DNoise(in vec3 pos)
{
	pos.z += 0.0f;
	vec3 p = floor(pos);
	vec3 f = fract(pos);
		 f = f * f * (3.0f - 2.0f * f);

	vec2 uv =  (p.xy + p.z * vec2(17.0f, 37.0f)) + f.xy;
	vec2 uv2 = (p.xy + (p.z + 1.0f) * vec2(17.0f, 37.0f)) + f.xy;
	vec2 coord =  (uv  + 0.5f) / 64.0f;
	vec2 coord2 = (uv2 + 0.5f) / 64.0f;
	vec3 xy1 = texture2D(noisetex, coord).xyz;
	vec3 xy2 = texture2D(noisetex, coord2).xyz;
	return mix(xy1, xy2, vec3(f.z));
}

vec3 Get3DNoiseNormal(in vec3 pos)
{
	float center = Get3DNoise(pos + vec3( 0.0f, 0.0f, 0.0f)).x * 2.0f - 1.0f;
	float left 	 = Get3DNoise(pos + vec3( 0.1f, 0.0f, 0.0f)).x * 2.0f - 1.0f;
	float up     = Get3DNoise(pos + vec3( 0.0f, 0.1f, 0.0f)).x * 2.0f - 1.0f;

	vec3 noiseNormal;
		 noiseNormal.x = center - left;
		 noiseNormal.y = center - up;

		 noiseNormal.x *= 0.2f;
		 noiseNormal.y *= 0.2f;

		 noiseNormal.b = sqrt(1.0f - noiseNormal.x * noiseNormal.x - noiseNormal.g * noiseNormal.g);
		 noiseNormal.b = 0.0f;

	return noiseNormal.xyz;
}


vec3 CalculateRainBump(in vec3 pos)
{

	

	pos.y += frameTimeCounter * 3.0f;
	pos.xz *= 1.0f;

	pos.y += Get3DNoise(pos.xyz * vec3(1.0f, 0.0f, 1.0f)).y * 2.0f;


	vec3 p = pos;
	vec3 noiseNormal = Get3DNoiseNormal(p);	p.y += 0.25f;
		 noiseNormal += Get3DNoiseNormal(p); p.y += 0.5f;
		 noiseNormal += Get3DNoiseNormal(p); p.y += 0.75f;
		 noiseNormal += Get3DNoiseNormal(p);
		 noiseNormal /= 4.0f;

	return Get3DNoiseNormal(pos).xyz;
}

float GetModulatedRainSpecular(in vec3 pos)
{
	pos.y += frameTimeCounter * 3.0f;
	pos.xz *= 0.5f;
	pos.y *= 0.2f;

	pos.y += Get3DNoise(pos.xyz * vec3(1.0f, 0.0f, 1.0f)).x * 2.0f;

	vec3 p = pos;

	float n = Get3DNoise(p).y;

	return n;
}

vec2 CalculateParallaxCoord(in vec2 coord, in vec3 viewVector)
{
	vec2 parallaxCoord = coord.st;
	const int maxSteps = 112;
	const vec3 stepSize = vec3(0.002f, 0.002f, 0.2f);

	float heightmap = TextureSmooth(normals, coord.st, 0, false).a;

	//if (viewVector.z < 0.0f)
	//{
		vec3 pCoord = vec3(0.0f, 0.0f, 1.0f);

		//make "pop out"
		//pCoord.st += (viewVector.xy * stepSize.xy) / (viewVector.z * stepSize.z);

		if (heightmap < 1.0f)
		{
			vec3 step = viewVector * stepSize;
			float distAngleWeight = ((distance * 0.6f) * (2.1f - viewVector.z)) / 4.0;
				 step *= distAngleWeight;

			float sampleHeight = TextureSmooth(normals, OffsetCoord(coord.st, pCoord.st, 0), 0, false).a;

			for (int i = 0; sampleHeight < pCoord.z && i < 24; ++i)
			{
				//if (heightmap < pCoord.z)
				pCoord.xy = mix(pCoord.xy, pCoord.xy + step.xy, clamp((pCoord.z - sampleHeight) / (stepSize.z * 1.75 * distAngleWeight / (-viewVector.z + 0.85)), 0.0, 1.0));
				pCoord.z += step.z;
				//pCoord += step;
				sampleHeight = TextureSmooth(normals, OffsetCoord(coord.st, pCoord.st, 0), 0, false).a;

			}


			parallaxCoord.xy = OffsetCoord(coord.st, pCoord.st, 0);
		}

	//}

	//parallaxCoord.xy = OffsetCoord(coord.st, viewVector.xy * (1.0f - heightmap) * 0.0045f, 0);

	return parallaxCoord;
}

vec2 CalculateShadowCoord(in vec2 coord, in vec3 viewVector)
{
	vec2 parallaxCoord = coord.st;
	const int maxSteps = 1122;
	const vec3 stepSize = vec3(0.001f, 0.001f, 0.1f);

	float heightmap = TextureSmooth(normals, coord.st, 0, false).a;

	//if (viewVector.z < 0.0f)
	//{
		vec3 pCoord = vec3(0.0f, 0.0f, 1.0f);

		//make "pop out"
		pCoord.st += (viewVector.xy * stepSize.xy) / (viewVector.z * stepSize.z);


		//viewVector.xy = - viewVector.xy;

		if (heightmap < 1.0f)
		{
			vec3 step = viewVector * stepSize * 0.25f;

			float sampleHeight = TextureSmooth(normals, OffsetCoord(coord.st, pCoord.st, 0), 0, false).a;

			for (int i = 0; sampleHeight < pCoord.z && i < 64; ++i)
			{
				//if (heightmap < pCoord.z)
				pCoord = mix(pCoord, pCoord + step, clamp((pCoord.z - sampleHeight) / (stepSize.z * 1.0 / (-viewVector.z + 0.01)), 0.0, 1.0));
				sampleHeight = TextureSmooth(normals, OffsetCoord(coord.st, pCoord.st, 0), 0, false).a;
				//pCoord += step;

			}


			parallaxCoord.xy = OffsetCoord(coord.st, pCoord.st, 0);
		}

	//}

	//parallaxCoord.xy = OffsetCoord(coord.st, viewVector.xy * (1.0f - heightmap) * 0.0045f, 0);

	return parallaxCoord;
}

void main() {	

	vec4 modelView = (gl_ModelViewMatrix * vertexPos);
		 // modelView.x *= aspectRatio;
		 // modelView.z *= 1.4f;
		 // modelView = gl_ProjectionMatrix * modelView;
		 // modelView.xyz /= modelView.w;

		 // modelView.z = -modelView.z;

	vec3 viewVector = normalize(tbnMatrix * modelView.xyz);
		 viewVector.x /= 2.0f;

		 viewVector = normalize(viewVector);
		 //viewVector = normalize(viewVector);
		 //viewVector.y *= 0.5f;
	vec3 shadowVector = normalize(tbnMatrix * -sunPosition.xyz);
		 //shadowVector.xy = -shadowVector.xy;

	vec2 parallaxCoord = texcoord.st;
		 //parallaxCoord = CalculateParallaxCoord(texcoord.st, viewVector);
	//vec2 shadowCoord = CalculateShadowCoord(parallaxCoord.st, shadowVector);

	float height = TextureSmooth(normals, parallaxCoord, 0, false).a;
	//float shadowHeight = TextureSmooth(normals, shadowCoord, 0, false).a;

	float sunlightVisibility = 1.0f;

	//if (shadowHeight > height + 0.00001f)
	//	sunlightVisibility = 0.0f;



	int tileResolution = TILE_RESOLUTION;
	float lod = 0.085f * tileResolution / 128;
	float mipLevel = distance * lod;
		  mipLevel += pow(clamp((1.0f + viewVector.z), 0.0f, 1.0f), 2.0f) * 3.5f * tileResolution / 128;
		  mipLevel = min(floor(mipLevel), 4.0f);
		
	vec4 spec = TextureSmooth(specular, parallaxCoord.st, int(mipLevel), false);
		 //spec.g += GetModulatedRainSpecular(worldPosition.xyz) * 0.8f + 0.0f;

	//store lightmap in auxilliary texture. r = torch light. g = lightning. b = sky light.
	vec4 lightmap = vec4(0.0f, 0.0f, 0.0f, 1.0f);
	
	//Separate lightmap types
	lightmap.r = clamp((lmcoord.s * 33.05f / 32.0f) - 1.05f / 32.0f, 0.0f, 1.0f);
	lightmap.b = clamp((lmcoord.t * 33.05f / 32.0f) - 1.05f / 32.0f, 0.0f, 1.0f);

	
	vec4 clr = color;
		 clr.rgb = clr.rgb / normalize(clr.rbg);
		 clr.rgb *= 0.5f;
		 clr.rgb = vec3(max(color.r, max(color.g, color.b)));

	float ao = (color.r + color.g + color.b) / 3.0f;

	float colorDiff = abs(color.r - color.g);
		  colorDiff += abs(color.r - color.b);
		  colorDiff += abs(color.g - color.b);

	if (colorDiff > 0.001f) {
		ao = 1.0f;
	}

	ao = pow(ao, 13.0f);

	 lightmap.b *= ao;

	 lightmap.r *= ao * 0.5f + 0.5f;

	 float wetfactor = clamp(lightmap.b - 0.9f, 0.0f, 0.1f) / 0.1f;

	lightmap.b = pow(lightmap.b, 1.0f);
	lightmap.r = pow(lightmap.r, 3.0f);

	
	
	
	vec4 frag2;
	
	if (distance < bump_distance) {
	
			vec3 bump = TextureSmooth(normals, parallaxCoord.st, int(mipLevel), false).rgb * 2.0f - 1.0f;
			
			float bumpmult = clamp(bump_distance * fademult - distance * fademult, 0.0f, 1.0f) * NORMAL_MAP_MAX_ANGLE;
	              bumpmult *= 1.0f - (pow(spec.g, 0.5f) * 0.9f * rainStrength * wetfactor);
				  
			bump = bump * vec3(bumpmult, bumpmult, bumpmult) + vec3(0.0f, 0.0f, 1.0f - bumpmult);

			//bump += CalculateRainBump(worldPosition.xyz);
			
			frag2 = vec4(bump * tbnMatrix * 0.5 + 0.5, 1.0);
			
	} else {
	
			frag2 = vec4((normal) * 0.5f + 0.5f, 1.0f);					
	}

	//Diffuse
	vec4 albedo = TextureSmooth(texture, parallaxCoord.st, int(mipLevel), true) * color;

		//sunlightVisibility *= clamp(dot(frag2.rgb * 2.0f - 1.0f, normalize(sunPosition.xyz)), 0.0f, 1.0f);

		//albedo.rgb *= sunlightVisibility * 0.8f + 0.2f;

		 //albedo.rgb *= TextureSmooth(normals, parallaxCoord.st, int(mipLevel), false).a;

		// vec3 noise = Get3DNoise(worldPosition.xyz);

		// albedo.rgb = noise.rgb;

		//Wet diffuse
		const float wetPow = 1.75f;
		
		float metallicMask = 0.0f;
		
		if (   abs(materialIDs - 20.0f) < 0.1f
			|| abs(materialIDs - 21.0f) < 0.1f
			|| abs(materialIDs - 22.0f) < 0.1f
			|| abs(materialIDs - 23.0f) < 0.1f) {
			metallicMask = 1.0f;
		}
		
		gl_FragData[0] = albedo;
		

	float mats_1 = materialIDs;
		  mats_1 += 0.1f;
	//Depth  
	gl_FragData[1] = vec4(mats_1/255.0f, lightmap.r, lightmap.b, 1.0f);

	//normal
	gl_FragData[2] = frag2;
		
	//specularity
	gl_FragData[3] = vec4(spec.r + spec.g * rainStrength * wetfactor, spec.b, 0.0f, 1.0f);	

}